#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

class CppMethods(object):
    @property
    def awkward(self):
        import awkward.cpp
        return awkward.cpp

    @property
    def JaggedArray(self):
        import awkward.cpp.array.jagged
        return awkward.cpp.array.jagged.JaggedArrayCpp



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import codecs
import json

import numpy

import awkward.array.base
import awkward.array.chunked
import awkward.array.indexed
import awkward.array.jagged
import awkward.array.masked
import awkward.array.objects
import awkward.array.table
import awkward.array.virtual
import awkward.type
import awkward.util

################################################################################ type conversions

def schema2type(schema):
    import pyarrow

    def recurse(tpe, nullable):
        if isinstance(tpe, pyarrow.lib.DictionaryType):
            out = recurse(tpe.dictionary.type, nullable)
            if nullable:
                return awkward.type.OptionType(out)
            else:
                return out

        elif isinstance(tpe, pyarrow.lib.StructType):
            out = None
            for i in range(tpe.num_children):
                x = awkward.type.ArrayType(tpe[i].name, recurse(tpe[i].type, tpe[i].nullable))
                if out is None:
                    out = x
                else:
                    out = out & x
            if nullable:
                return awkward.type.OptionType(out)
            else:
                return out

        elif isinstance(tpe, pyarrow.lib.ListType):
            out = awkward.type.ArrayType(float("inf"), recurse(tpe.value_type, nullable))
            if nullable:
                return awkward.type.OptionType(out)
            else:
                return out

        elif isinstance(tpe, pyarrow.lib.UnionType):
            out = None
            for i in range(tpe.num_children):
                x = recurse(tpe[i].type, nullable)
                if out is None:
                    out = x
                else:
                    out = out | x
            if nullable:
                return awkward.type.OptionType(out)
            else:
                return out

        elif tpe == pyarrow.string():
            if nullable:
                return awkward.type.OptionType(str)
            else:
                return str

        elif tpe == pyarrow.binary():
            if nullable:
                return awkward.type.OptionType(bytes)
            else:
                return bytes

        elif tpe == pyarrow.bool_():
            out = awkward.numpy.dtype(bool)
            if nullable:
                return awkward.type.OptionType(out)
            else:
                return out

        elif isinstance(tpe, pyarrow.lib.DataType):
            if nullable:
                return awkward.type.OptionType(tpe.to_pandas_dtype())
            else:
                return tpe.to_pandas_dtype()

        else:
            raise NotImplementedError(repr(tpe))

    out = None
    for name in schema.names:
        field = schema.field_by_name(name)
        mytype = awkward.type.ArrayType(name, recurse(field.type, field.nullable))
        if out is None:
            out = mytype
        else:
            out = out & mytype

    return out

################################################################################ value conversions

def toarrow(obj):
    import pyarrow

    def recurse(obj, mask):
        if isinstance(obj, numpy.ndarray):
            return pyarrow.array(obj, mask=mask)

        elif isinstance(obj, awkward.array.chunked.ChunkedArray):   # includes AppendableArray
            raise TypeError("only top-level ChunkedArrays can be converted to Arrow (as RecordBatches)")

        elif isinstance(obj, awkward.array.indexed.IndexedArray):
            if mask is None:
                return pyarrow.DictionaryArray.from_arrays(obj.index, recurse(obj.content, mask))
            else:
                return recurse(obj.content[obj.index], mask)

        elif isinstance(obj, awkward.array.indexed.SparseArray):
            return recurse(obj.dense, mask)

        elif isinstance(obj, awkward.array.jagged.JaggedArray):
            obj = obj.compact()
            if mask is not None:
                mask = obj.tojagged(mask).flatten()
            return pyarrow.ListArray.from_arrays(obj.offsets, recurse(obj.content, mask))

        elif isinstance(obj, awkward.array.masked.IndexedMaskedArray):
            thismask = obj.boolmask(maskedwhen=True)
            if mask is not None:
                thismask = mask | thismask
            if len(obj.content) == 0:
                content = obj.numpy.empty(len(obj.mask), dtype=obj.DEFAULTTYPE)
            else:
                content = obj.content[obj.mask]
            return recurse(content, thismask)

        elif isinstance(obj, awkward.array.masked.MaskedArray):   # includes BitMaskedArray
            thismask = obj.boolmask(maskedwhen=True)
            if mask is not None:
                thismask = mask | thismask
            return recurse(obj.content, thismask)

        elif isinstance(obj, awkward.array.objects.StringArray):
            # # FIXME: BinaryArray.from_buffers is not implemented in pyarrow yet.
            # if obj.encoding is None:
            #     convert = lambda length, offsets, content: pyarrow.BinaryArray.from_buffers(pyarrow.binary(), length, [None, offsets, content])
            # elif codecs.lookup(obj.encoding) is codecs.lookup("utf-8"):
            #     convert = lambda length, offsets, content: pyarrow.StringArray.from_buffers(length, offsets, content)
            # else:
            #     raise ValueError("only encoding=None or encoding='utf-8' can be converted to Arrow")
            convert = lambda length, offsets, content: pyarrow.StringArray.from_buffers(length, offsets, content)

            obj = obj.compact()
            offsets = obj.offsets
            if offsets.dtype != numpy.dtype(numpy.int32):
                offsets = offsets.astype(numpy.int32)

            return convert(len(offsets) - 1, pyarrow.py_buffer(offsets), pyarrow.py_buffer(obj.content))

        elif isinstance(obj, awkward.array.objects.ObjectArray):
            # throw away Python object interpretation, which Arrow can't handle while being multilingual
            return recurse(obj.content, mask)

        elif isinstance(obj, awkward.array.table.Table):
            return pyarrow.StructArray.from_arrays([recurse(x, mask) for x in obj.contents.values()], list(obj.contents))

        elif isinstance(obj, awkward.array.union.UnionArray):
            contents = []
            for i, x in enumerate(obj.contents):
                if mask is None:
                    thismask = None
                else:
                    thistags = (obj.tags == i)
                    thismask = obj.numpy.empty(len(x), dtype=obj.MASKTYPE)
                    thismask[obj.index[thistags]] = mask[thistags]    # hmm... obj.index could have repeats; the Arrow mask in that case would not be well-defined...
                contents.append(recurse(x, thismask))

            return pyarrow.UnionArray.from_dense(pyarrow.array(obj.tags.astype(numpy.int8)), pyarrow.array(obj.index.astype(numpy.int32)), contents)

        elif isinstance(obj, awkward.array.virtual.VirtualArray):
            return recurse(obj.array, mask)

        else:
            raise TypeError("cannot convert type {0} to Arrow".format(type(obj)))

    if isinstance(obj, awkward.array.chunked.ChunkedArray):   # includes AppendableArray
        batches = []
        for chunk in obj.chunks:
            arr = toarrow(chunk)
            if isinstance(arr, pyarrow.Table):
                batches.extend(arr.to_batches())
            else:
                batches.append(pyarrow.RecordBatch.from_arrays([arr], [""]))
        return pyarrow.Table.from_batches(batches)

    elif isinstance(obj, awkward.array.masked.IndexedMaskedArray) and isinstance(obj.content, awkward.array.table.Table):
        mask = obj.boolmask(maskedwhen=True)
        if len(obj.content) == 0:
            content = obj.numpy.empty(len(obj.mask), dtype=obj.DEFAULTTYPE)
        else:
            content = obj.content[obj.mask]
        return pyarrow.Table.from_batches([pyarrow.RecordBatch.from_arrays([recurse(x, mask) for x in obj.content.contents.values()], list(obj.content.contents))])

    elif isinstance(obj, awkward.array.masked.MaskedArray) and isinstance(obj.content, awkward.array.table.Table):   # includes BitMaskedArray
        mask = obj.boolmask(maskedwhen=True)
        return pyarrow.Table.from_batches([pyarrow.RecordBatch.from_arrays([recurse(x, mask) for x in obj.content.contents.values()], list(obj.content.contents))])

    elif isinstance(obj, awkward.array.table.Table):
        return pyarrow.Table.from_batches([pyarrow.RecordBatch.from_arrays([recurse(x, None) for x in obj.contents.values()], list(obj.contents))])

    else:
        return recurse(obj, None)

def fromarrow(obj, awkwardlib=None):
    import pyarrow
    awkwardlib = awkward.util.awkwardlib(awkwardlib)
    ARROW_BITMASKTYPE = awkwardlib.numpy.uint8
    ARROW_INDEXTYPE = awkwardlib.numpy.int32
    ARROW_TAGTYPE = awkwardlib.numpy.uint8
    ARROW_CHARTYPE = awkwardlib.numpy.uint8

    def popbuffers(array, tpe, buffers, length):
        if isinstance(tpe, pyarrow.lib.DictionaryType):
            index = popbuffers(None if array is None else array.indices, tpe.index_type, buffers, length)
            if hasattr(tpe, "dictionary"):
                content = fromarrow(tpe.dictionary)
            elif array is not None:
                content = fromarrow(array.dictionary)
            else:
                raise NotImplementedError("no way to access Arrow dictionary inside of UnionArray")
            if isinstance(index, awkwardlib.BitMaskedArray):
                return awkwardlib.BitMaskedArray(index.mask, awkwardlib.IndexedArray(index.content, content), maskedwhen=index.maskedwhen, lsborder=index.lsborder)
            else:
                return awkwardlib.IndexedArray(index, content)

        elif isinstance(tpe, pyarrow.lib.StructType):
            assert getattr(tpe, "num_buffers", 1) == 1
            mask = buffers.pop(0)
            pairs = []
            for i in range(tpe.num_children):
                pairs.append((tpe[i].name, popbuffers(None if array is None else array.field(tpe[i].name), tpe[i].type, buffers, length)))
            out = awkwardlib.Table.frompairs(pairs, 0)   # FIXME: better rowstart
            if mask is not None:
                mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
                return awkwardlib.BitMaskedArray(mask, out, maskedwhen=False, lsborder=True)
            else:
                return out

        elif isinstance(tpe, pyarrow.lib.ListType):
            assert getattr(tpe, "num_buffers", 2) == 2
            mask = buffers.pop(0)
            offsets = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_INDEXTYPE)[:length + 1]
            content = popbuffers(None if array is None else array.flatten(), tpe.value_type, buffers, offsets[-1])
            out = awkwardlib.JaggedArray.fromoffsets(offsets, content)
            if mask is not None:
                mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
                return awkwardlib.BitMaskedArray(mask, out, maskedwhen=False, lsborder=True)
            else:
                return out

        elif isinstance(tpe, pyarrow.lib.UnionType) and tpe.mode == "sparse":
            assert getattr(tpe, "num_buffers", 3) == 3
            mask = buffers.pop(0)
            tags = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_TAGTYPE)[:length]
            assert buffers.pop(0) is None
            index = awkwardlib.numpy.arange(len(tags), dtype=ARROW_INDEXTYPE)
            contents = []
            for i in range(tpe.num_children):
                try:
                    sublength = index[tags == i][-1] + 1
                except IndexError:
                    sublength = 0
                contents.append(popbuffers(None, tpe[i].type, buffers, sublength))
            for i in range(len(contents)):
                these = index[tags == i]
                if len(these) == 0:
                    contents[i] = contents[i][0:0]
                else:
                    contents[i] = contents[i][: these[-1] + 1]
            out = awkwardlib.UnionArray(tags, index, contents)
            if mask is not None:
                mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
                return awkwardlib.BitMaskedArray(mask, out, maskedwhen=False, lsborder=True)
            else:
                return out

        elif isinstance(tpe, pyarrow.lib.UnionType) and tpe.mode == "dense":
            assert getattr(tpe, "num_buffers", 3) == 3
            mask = buffers.pop(0)
            tags = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_TAGTYPE)[:length]
            index = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_INDEXTYPE)[:length]
            contents = []
            for i in range(tpe.num_children):
                try:
                    sublength = index[tags == i].max() + 1
                except ValueError:
                    sublength = 0
                contents.append(popbuffers(None, tpe[i].type, buffers, sublength))
            for i in range(len(contents)):
                these = index[tags == i]
                if len(these) == 0:
                    contents[i] = contents[i][0:0]
                else:
                    contents[i] = contents[i][: these.max() + 1]
            out = awkwardlib.UnionArray(tags, index, contents)
            if mask is not None:
                mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
                return awkwardlib.BitMaskedArray(mask, out, maskedwhen=False, lsborder=True)
            else:
                return out

        elif tpe == pyarrow.string():
            assert getattr(tpe, "num_buffers", 3) == 3
            mask = buffers.pop(0)
            offsets = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_INDEXTYPE)[:length + 1]
            content = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_CHARTYPE)[:offsets[-1]]
            out = awkwardlib.StringArray.fromoffsets(offsets, content[:offsets[-1]], encoding="utf-8")
            if mask is not None:
                mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
                return awkwardlib.BitMaskedArray(mask, out, maskedwhen=False, lsborder=True)
            else:
                return out

        elif tpe == pyarrow.binary():
            assert getattr(tpe, "num_buffers", 3) == 3
            mask = buffers.pop(0)
            offsets = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_INDEXTYPE)[:length + 1]
            content = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_CHARTYPE)[:offsets[-1]]
            out = awkwardlib.StringArray.fromoffsets(offsets, content[:offsets[-1]], encoding=None)
            if mask is not None:
                mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
                return awkwardlib.BitMaskedArray(mask, out, maskedwhen=False, lsborder=True)
            else:
                return out

        elif tpe == pyarrow.bool_():
            assert getattr(tpe, "num_buffers", 2) == 2
            mask = buffers.pop(0)
            out = awkwardlib.numpy.unpackbits(awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=ARROW_CHARTYPE)).view(awkwardlib.MaskedArray.BOOLTYPE)
            out = out.reshape(-1, 8)[:,::-1].reshape(-1)[:length]    # lsborder=True
            if mask is not None:
                mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
                return awkwardlib.BitMaskedArray(mask, out, maskedwhen=False, lsborder=True)
            else:
                return out

        elif isinstance(tpe, pyarrow.lib.DataType):
            assert getattr(tpe, "num_buffers", 2) == 2
            mask = buffers.pop(0)
            out = awkwardlib.numpy.frombuffer(buffers.pop(0), dtype=tpe.to_pandas_dtype())[:length]
            if mask is not None:
                mask = awkwardlib.numpy.frombuffer(mask, dtype=ARROW_BITMASKTYPE)
                return awkwardlib.BitMaskedArray(mask, out, maskedwhen=False, lsborder=True)
            else:
                return out

        else:
            raise NotImplementedError(repr(tpe))

    if isinstance(obj, pyarrow.lib.Array):
        buffers = obj.buffers()
        out = popbuffers(obj, obj.type, buffers, len(obj))
        assert len(buffers) == 0
        return out

    elif isinstance(obj, pyarrow.lib.ChunkedArray):
        chunks = [x for x in obj.chunks if len(x) > 0]
        if len(chunks) == 1:
            return chunks[0]
        else:
            return awkwardlib.ChunkedArray([fromarrow(x) for x in chunks], chunksizes=[len(x) for x in chunks])

    elif isinstance(obj, pyarrow.lib.RecordBatch):
        out = awkwardlib.Table()
        for n, x in zip(obj.schema.names, obj.columns):
            out[n] = fromarrow(x)
        return out

    elif isinstance(obj, pyarrow.lib.Table):
        chunks = []
        chunksizes = []
        for batch in obj.to_batches():
            chunk = fromarrow(batch)
            if len(chunk) > 0:
                chunks.append(chunk)
                chunksizes.append(len(chunk))
        if len(chunks) == 1:
            return chunks[0]
        else:
            return awkwardlib.ChunkedArray(chunks, chunksizes=chunksizes)

    else:
        raise NotImplementedError(type(obj))

################################################################################ Parquet file handling

def toparquet(where, obj, **options):
    import pyarrow.parquet

    options["where"] = where

    def convert(obj, message):
        if isinstance(obj, (awkward.array.base.AwkwardArray, numpy.ndarray)):
            out = toarrow(obj)
            if isinstance(out, pyarrow.Table):
                return out
            else:
                return pyarrow.Table.from_batches([pyarrow.RecordBatch.from_arrays([out], [""])])
        else:
            raise TypeError(message)

    if isinstance(obj, awkward.array.chunked.ChunkedArray):
        obj = iter(obj.chunks)
        try:
            awkitem = next(obj)
        except StopIteration:
            raise ValueError("iterable is empty")

        arritem = convert(awkitem, None)
        if "schema" not in options:
            options["schema"] = arritem.schema
        writer = pyarrow.parquet.ParquetWriter(**options)
        writer.write_table(arritem)

        try:
            while True:
                try:
                    awkitem = next(obj)
                except StopIteration:
                    break
                else:
                    writer.write_table(convert(awkitem, None))
        finally:
            writer.close()

    elif isinstance(obj, (awkward.array.base.AwkwardArray, numpy.ndarray)):
        arritem = convert(obj, None)
        options["schema"] = arritem.schema
        writer = pyarrow.parquet.ParquetWriter(**options)
        writer.write_table(arritem)
        writer.close()

    else:
        try:
            obj = iter(obj)
        except TypeError:
            raise TypeError("cannot write {0} to Parquet file".format(type(obj)))
        try:
            awkitem = next(obj)
        except StopIteration:
            raise ValueError("iterable is empty")

        arritem = convert(awkitem, "cannot write iterator of {0} to Parquet file".format(type(awkitem)))
        if "schema" not in options:
            options["schema"] = arritem.schema
        writer = pyarrow.parquet.ParquetWriter(**options)
        writer.write_table(arritem)

        try:
            while True:
                try:
                    awkitem = next(obj)
                except StopIteration:
                    break
                else:
                    writer.write_table(convert(awkitem, "cannot write iterator of {0} to Parquet file".format(type(awkitem))))
        finally:
            writer.close()

class _ParquetFile(object):
    def __init__(self, file, metadata=None, common_metadata=None):
        self.file = file
        self.metadata = metadata
        self.common_metadata = common_metadata
        self._init()

    def _init(self):
        import pyarrow.parquet
        self.parquetfile = pyarrow.parquet.ParquetFile(self.file, metadata=self.metadata, common_metadata=self.common_metadata)
        self.type = schema2type(self.parquetfile.schema.to_arrow_schema())

    def __getstate__(self):
        return {"file": self.file, "metadata": self.metadata, "common_metadata": self.common_metadata}

    def __setstate__(self, state):
        self.file = state["file"]
        self.metadata = state["metadata"]
        self.common_metadata = state["common_metadata"]
        self._init()

    def __call__(self, rowgroup, column):
        return fromarrow(self.parquetfile.read_row_group(rowgroup, columns=[column]))[column]

    def tojson(self):
        json.dumps([self.file, self.metadata, self.common_metadata])
        return {"file": self.file, "metadata": self.metadata, "common_metadata": self.common_metadata}

    @classmethod
    def fromjson(cls, state):
        return cls(state["file"], metadata=state["metadata"], common_metadata=state["common_metadata"])

def fromparquet(file, awkwardlib=None, cache=None, persistvirtual=False, metadata=None, common_metadata=None):
    awkwardlib = awkward.util.awkwardlib(awkwardlib)
    parquetfile = _ParquetFile(file, metadata=metadata, common_metadata=common_metadata)
    columns = parquetfile.type.columns

    chunks = []
    chunksizes = []
    for i in range(parquetfile.parquetfile.num_row_groups):
        numrows = parquetfile.parquetfile.metadata.row_group(i).num_rows
        if numrows > 0:
            if columns == [""]:
                chunk = awkwardlib.VirtualArray(parquetfile, (i, ""), cache=cache, type=awkwardlib.type.ArrayType(numrows, parquetfile.type[""]), persistvirtual=persistvirtual)
            else:
                chunk = awkwardlib.Table()
                for n in columns:
                    q = awkwardlib.VirtualArray(parquetfile, (i, n), cache=cache, type=awkwardlib.type.ArrayType(numrows, parquetfile.type[n]), persistvirtual=persistvirtual)
                    chunk.contents[n] = q

            chunks.append(chunk)
            chunksizes.append(numrows)

    return awkwardlib.ChunkedArray(chunks, chunksizes)



#!/usr/bin/env python

# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-array/blob/master/LICENSE

import unittest

import pytest
import numpy

awkward_numba = pytest.importorskip("awkward.numba")

class Test(unittest.TestCase):
    def runTest(self):
        pass

    def test_awkward_numba_init(self):
        a = awkward_numba.JaggedArray([0, 3, 3, 5], [3, 3, 5, 10], [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
        assert isinstance(a, awkward_numba.array.jagged.JaggedArrayNumba)
        assert a.JaggedArray is awkward_numba.array.jagged.JaggedArrayNumba

        assert isinstance(a[2:], awkward_numba.array.jagged.JaggedArrayNumba)
        assert isinstance(a[[False, True, True, False]], awkward_numba.array.jagged.JaggedArrayNumba)
        assert isinstance(a[[1, 2]], awkward_numba.array.jagged.JaggedArrayNumba)
        assert isinstance(a + 100, awkward_numba.array.jagged.JaggedArrayNumba)
        assert isinstance(numpy.sin(a), awkward_numba.array.jagged.JaggedArrayNumba)
        assert isinstance(a.cross(a), awkward_numba.array.jagged.JaggedArrayNumba)
        assert isinstance(a.pairs(), awkward_numba.array.jagged.JaggedArrayNumba)
        assert isinstance(a.distincts(), awkward_numba.array.jagged.JaggedArrayNumba)
        assert isinstance(a.cross(a, nested=True), awkward_numba.array.jagged.JaggedArrayNumba)
        assert isinstance(a.pairs(nested=True), awkward_numba.array.jagged.JaggedArrayNumba)
        assert isinstance(a.distincts(nested=True), awkward_numba.array.jagged.JaggedArrayNumba)
        assert isinstance(a.concatenate([a]), awkward_numba.array.jagged.JaggedArrayNumba)

        b = awkward_numba.JaggedArray([1, 4, 4, 6], [4, 4, 6, 11], [999, 0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
        assert isinstance(a + b, awkward_numba.array.jagged.JaggedArrayNumba)

    def test_awkward_numba_argmin(self):
        a = awkward_numba.JaggedArray([0, 3, 3, 5], [3, 3, 5, 10], [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
        assert isinstance(a.argmin(), awkward_numba.array.jagged.JaggedArrayNumba)
        assert a.argmin().tolist() == [[0], [], [0], [0]]

        a = awkward_numba.JaggedArray([[0, 3], [3, 5]], [[3, 3], [5, 10]], [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
        assert isinstance(a.argmin(), awkward_numba.array.jagged.JaggedArrayNumba)
        assert a.argmin().tolist() == [[[0], []], [[0], [0]]]

    def test_awkward_numba_argmax(self):
        a = awkward_numba.JaggedArray([0, 3, 3, 5], [3, 3, 5, 10], [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
        assert isinstance(a.argmin(), awkward_numba.array.jagged.JaggedArrayNumba)
        assert a.argmax().tolist() == [[2], [], [1], [4]]

        a = awkward_numba.JaggedArray([[0, 3], [3, 5]], [[3, 3], [5, 10]], [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
        assert isinstance(a.argmin(), awkward_numba.array.jagged.JaggedArrayNumba)
        assert a.argmax().tolist() == [[[2], []], [[1], [4]]]



#!/usr/bin/env python

import pandas
import awkward
import pyarrow

exoplanets = pandas.read_csv("/home/pivarski/talks/2019-05-28-lpchats-numpy-uproot-awkward/data/nasa-exoplanets.csv")
exoplanets.index = pandas.MultiIndex.from_arrays([exoplanets["pl_hostname"], exoplanets["pl_letter"]])
exoplanets.index.names = ["star", "planet"]

df = exoplanets[["ra", "dec", "st_dist", "st_mass", "st_rad", "pl_orbsmax", "pl_orbeccen", "pl_orbper", "pl_bmassj", "pl_radj"]]
df.columns = pandas.MultiIndex.from_arrays([["star"] * 5 + ["planet"] * 5,
    ["right asc. (deg)", "declination (deg)", "distance (pc)", "mass (solar)", "radius (solar)", "orbit (AU)", "eccen.", "period (days)", "mass (Jupiter)", "radius (Jupiter)"]])

stardicts = []
for (starname, planetname), row in df.iterrows():
    if len(stardicts) == 0 or stardicts[-1]["name"] != starname:
        stardicts.append({"name": starname,
                          "ra": row["star", "right asc. (deg)"],
                          "dec": row["star", "declination (deg)"],
                          "dist": row["star", "distance (pc)"],
                          "mass": row["star", "mass (solar)"],
                          "radius": row["star", "radius (solar)"],
                          "planets": []})
    stardicts[-1]["planets"].append({"name": planetname,
                                     "orbit": row["planet", "orbit (AU)"],
                                     "eccen": row["planet", "eccen."],
                                     "period": row["planet", "period (days)"],
                                     "mass": row["planet", "mass (Jupiter)"],
                                     "radius": row["planet", "radius (Jupiter)"]})

stars = awkward.fromiter(stardicts)
arrowstars = awkward.toarrow(stars)

with open("exoplanets.arrow", "wb") as sink:
    writer = pyarrow.RecordBatchFileWriter(sink, arrowstars.schema)
    for batch in arrowstars.to_batches():
        writer.write_batch(batch)
    writer.close()

# to read it back again:
awkward.fromarrow(pyarrow.ipc.open_file(open("exoplanets.arrow", "rb")).get_batch(0))



import time

import numpy
import numba

import awkward
import awkward.numba

num_muons = numpy.random.poisson(1.5, 100000)
pt_muons  = awkward.JaggedArray.fromcounts(num_muons, numpy.random.exponential(10, num_muons.sum()) + 50)
eta_muons = awkward.JaggedArray.fromcounts(num_muons, numpy.random.normal(0, 1, num_muons.sum()))
phi_muons = awkward.JaggedArray.fromcounts(num_muons, numpy.random.uniform(-numpy.pi, numpy.pi, num_muons.sum()))

num_jets = numpy.random.poisson(3.5, 100000)
pt_jets  = awkward.JaggedArray.fromcounts(num_jets, numpy.random.exponential(10, num_jets.sum()) + 50)
eta_jets = awkward.JaggedArray.fromcounts(num_jets, numpy.random.normal(0, 1, num_jets.sum()))
phi_jets = awkward.JaggedArray.fromcounts(num_jets, numpy.random.uniform(-numpy.pi, numpy.pi, num_jets.sum()))

####################################################### for-loopy

def run_python(num_muons, pt_muons, eta_muons, phi_muons, num_jets, pt_jets, eta_jets, phi_jets):
    offsets = numpy.empty(len(num_muons) + 1, numpy.int64)
    content = numpy.empty((num_muons * num_jets).sum())
    offsets[0] = 0
    for i in range(len(num_muons)):
        offsets[i + 1] = offsets[i]
        for muoni in range(num_muons[i]):
            pt1  = pt_muons[i][muoni]
            eta1 = eta_muons[i][muoni]
            phi1 = phi_muons[i][muoni]
            for jeti in range(num_jets[i]):
                pt2  = pt_jets[i][jeti]
                eta2 = eta_jets[i][jeti]
                phi2 = phi_jets[i][jeti]
                content[offsets[i + 1]] = numpy.sqrt(2*pt1*pt2*(numpy.cosh(eta1 - eta2) - numpy.cos(phi1 - phi2)))
                offsets[i + 1] += 1
    return awkward.JaggedArray(offsets[:-1], offsets[1:], content)

starttime = time.time()
mass1 = run_python(num_muons, pt_muons, eta_muons, phi_muons, num_jets, pt_jets, eta_jets, phi_jets)
print("Python", time.time() - starttime)

####################################################### Numpythonic

def run_numpy(num_muons, pt_muons, eta_muons, phi_muons, num_jets, pt_jets, eta_jets, phi_jets):
    def unzip(pairs):
        return pairs.i0, pairs.i1

    pt1, pt2   = unzip(pt_muons.cross(pt_jets))
    eta1, eta2 = unzip(eta_muons.cross(eta_jets))
    phi1, phi2 = unzip(phi_muons.cross(phi_jets))

    return numpy.sqrt(2*pt1*pt2*(numpy.cosh(eta1 - eta2) - numpy.cos(phi1 - phi2)))

for i in range(5):
    starttime = time.time()
    mass2 = run_numpy(num_muons, pt_muons, eta_muons, phi_muons, num_jets, pt_jets, eta_jets, phi_jets)
    print("Numpy", time.time() - starttime)

####################################################### with Numba

run_numba = numba.jit(nopython=True)(run_python)

for i in range(5):
    starttime = time.time()
    mass3 = run_numba(num_muons, pt_muons, eta_muons, phi_muons, num_jets, pt_jets, eta_jets, phi_jets)
    print("Numba", time.time() - starttime)



#!/usr/bin/env python

# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy
import awkward

def spread_advanced(starts, stops, advanced):
    if advanced is None:
        return advanced
    else:
        assert len(starts) == len(stops) == len(advanced)
        counts = stops - starts
        nextadvanced = numpy.full(counts.sum(), 999, int)
        k = 0
        for i in range(len(counts)):
            length = counts[i]
            nextadvanced[k : k + length] = advanced[i]
            k += length
        return nextadvanced

def getitem_integer(array, head, tail, advanced):
    index = numpy.full(len(array.starts), 999, int)
    for i in range(len(array.starts)):
        j = array.starts[i] + head
        if j >= array.stops[i]:
            raise ValueError("integer index is beyond the range of one of the JaggedArray.starts-JaggedArray.stops pairs")
        index[i] = j

    next = getitem_next(array.content[index], tail, advanced)    
    return next

def getitem_slice2(array, head, tail, advanced):
    if (head.start is None or head.start == 0) and head.stop is None:
        next = getitem_next(array.content, tail, advanced)
        return awkward.JaggedArray(array.starts, array.stops, next)

    starts = numpy.full(len(array.starts), 999, int)
    stops = numpy.full(len(array.starts), 999, int)
    for i in range(len(array.starts)):
        length = array.stops[i] - array.starts[i]
        a, b = head.start, head.stop

        if a is None:
            a = 0
        elif a < 0:
            a += length
        if b is None:
            b = length
        elif b < 0:
            b += length

        if b <= a:
            a, b = 0, 0
        if a < 0:
            a = 0
        elif a > length:
            a = length
        if b < 0:
            b = 0
        elif b > length:
            b = length

        starts[i] = array.starts[i] + a
        stops[i] = array.starts[i] + b

    next = getitem_next(array.content, tail, advanced)
    return awkward.JaggedArray(starts, stops, next)

def getitem_slice3(array, head, tail, advanced):
    if head.step == 0:
        raise ValueError

    offsets = numpy.full(len(array.starts) + 1, 999, int)
    offsets[0] = 0
    index = numpy.full(len(array.content), 999, int)  # too big, but okay
    k = 0
    for i in range(len(array.starts)):
        length = array.stops[i] - array.starts[i]
        a, b, c = head.start, head.stop, head.step
        if c is None:
            c = 1

        if a is None and c > 0:
            a = 0
        elif a is None:
            a = length - 1
        elif a < 0:
            a += length

        if b is None and c > 0:
            b = length
        elif b is None:
            b = -1
        elif b < 0:
            b += length

        if c > 0:
            if b <= a:
                a, b = 0, 0
            if a < 0:
                a = 0
            elif a > length:
                a = length
            if b < 0:
                b = 0
            elif b > length:
                b = length
        else:
            if a <= b:
                a, b = 0, 0
            if a < -1:
                a = -1
            elif a >= length:
                a = length - 1
            if b < -1:
                b = -1
            elif b >= length:
                b = length - 1

        for j in range(a, b, c):
            index[k] = array.starts[i] + j
            k += 1
        offsets[i + 1] = k

    starts = offsets[:-1]
    stops = offsets[1:]
    next = getitem_next(array.content[index[:k]], tail, spread_advanced(starts, stops, advanced))
    return awkward.JaggedArray(starts, stops, next)

def getitem_intarray_none(array, head, tail, advanced):
    offsets = numpy.full(len(array.starts) + 1, 999, int)
    offsets[0] = 0
    index = numpy.full(len(head)*len(array.starts), 999, int)
    nextadvanced = numpy.full(len(index), 999, int)

    k = 0
    for i in range(len(array.starts)):
        length = array.stops[i] - array.starts[i]

        for j in range(len(head)):
            norm = head[j]
            if norm < 0:
                norm += length
            if norm < 0 or norm >= length:
                raise IndexError("advanced index is out of bounds in JaggedArray")
            index[k] = array.starts[i] + norm
            nextadvanced[k] = j
            k += 1
        offsets[i + 1] = k

    starts = offsets[:-1]
    stops = offsets[1:]
    next = getitem_next(array.content[index], tail, nextadvanced)
    return awkward.JaggedArray(starts, stops, next)

def getitem_intarray_some(array, head, tail, advanced):
    index = numpy.full(len(array.starts), 999, int)
    nextadvanced = numpy.full(len(index), 999, int)

    for i in range(len(advanced)):
        length = array.stops[i] - array.starts[i]
        if advanced[i] >= len(head):
            raise IndexError("advanced index lengths do not match")
        norm = head[advanced[i]]
        if norm < 0:
            norm += length
        if norm < 0 or norm >= length:
            raise IndexError("advanced index is out of bounds in JaggedArray")
        index[i] = array.starts[i] + norm
        nextadvanced[i] = i

    next = getitem_next(array.content[index], tail, nextadvanced)
    return next

def getitem_next(array, where, advanced):
    if len(where) == 0:
        return array
    if isinstance(array, numpy.ndarray):
        return array[where]
    
    head = where[0]
    tail = where[1:]
    if isinstance(head, int):
        return getitem_integer(array, head, tail, advanced)

    elif isinstance(head, slice) and (head.step is None or head.step == 1) and advanced is None and not any(isinstance(x, numpy.ndarray) for x in where):
        return getitem_slice2(array, head, tail, advanced)

    elif isinstance(head, slice):
        return getitem_slice3(array, head, tail, advanced)

    elif isinstance(head, numpy.ndarray) and issubclass(head.dtype.type, numpy.integer):
        if advanced is None:
            return getitem_intarray_none(array, head, tail, advanced)
        else:
            return getitem_intarray_some(array, head, tail, advanced)

    else:
        raise NotImplementedError(head)

def getitem_enter(array, where):
    if len(where) == 0:
        return array

    arraylen = 0
    for x in where:
        if isinstance(x, numpy.ndarray) and len(x.shape) == 1:
            if issubclass(x.dtype.type, (numpy.bool_, numpy.bool)):
                arraylen = max(arraylen, numpy.count_nonzero(x))
            else:
                arraylen = max(arraylen, len(x))

    newwhere = []
    for x in where:
        if isinstance(x, numpy.ndarray) and len(x.shape) == 1 and issubclass(x.dtype.type, (numpy.bool, numpy.bool_)):
            newwhere.append(numpy.nonzero(x)[0])
        elif isinstance(x, int) and arraylen != 0:
            newwhere.append(numpy.full(arraylen, x, int))
        elif isinstance(x, numpy.ndarray) and x.shape == (1,):
            newwhere.append(numpy.full(arraylen, x, int))
        else:
            newwhere.append(x)

    fake = getitem_next(awkward.JaggedArray([0], [len(array)], array), newwhere, None)
    if isinstance(fake, numpy.ndarray):
        return fake[0]
    else:
        return fake.content[fake.starts[0]:fake.stops[-1]]

slices = [2, slice(None), slice(2, 4), slice(1, None, 2), slice(None, None, -1), numpy.array([2, 0, 0]), numpy.array([3, 1, 2]), numpy.array([True, False, True, True]), numpy.array([True, True, True, False])]

a = numpy.arange(4**4).reshape(4, 4, 4, 4)
a2 = awkward.fromiter(a)

for x in slices:
    print(x)
    assert a[x,].tolist() == getitem_enter(a2, (x,)).tolist()
    for y in slices:
        print(x, y)
        assert a[x, y].tolist() == getitem_enter(a2, (x, y)).tolist()
        for z in slices:
            print(x, y, z)
            assert a[x, y, z].tolist() == getitem_enter(a2, (x, y, z)).tolist()

tmp = a2.content.content.content.tolist()
tmp.insert(4, 999)
a2.content.content.stops = a2.content.content.stops.copy()
a2.content.content.starts[1:] += 1
a2.content.content.stops[1:] += 1
a2.content.content.content = tmp

for x in slices:
    print(x)
    assert a[x,].tolist() == getitem_enter(a2, (x,)).tolist()
    for y in slices:
        print(x, y)
        assert a[x, y].tolist() == getitem_enter(a2, (x, y)).tolist()
        for z in slices:
            print(x, y, z)
            assert a[x, y, z].tolist() == getitem_enter(a2, (x, y, z)).tolist()

a = numpy.arange(4**3).reshape(4, 4, 4)
a2 = awkward.fromiter(a)
for x in slices:
    print(x)
    assert a[x,].tolist() == getitem_enter(a2, (x,)).tolist()
    for y in slices:
        print(x, y)
        assert a[x, y].tolist() == getitem_enter(a2, (x, y)).tolist()
        for z in slices:
            print(x, y, z)
            assert a[x, y, z].tolist() == getitem_enter(a2, (x, y, z)).tolist()
